DEFAULT_CIPHERS = ("ECDH+AESGCM:DH+AESGCM:ECDH+AES256:DH+AES256:ECDH+AES128:"
                   "DH+AES:ECDH+HIGH:DH+HIGH:ECDH+3DES:DH+3DES:RSA+AESGCM:"
                   "RSA+AES:RSA+HIGH:RSA+3DES:!aNULL:!eNULL:!MD5")
